/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Hash table functions
 *
 * Marcio Oliveira -- 2002
 */

#include <stdlib.h>
#include <error.h>
#include <hash.h>

hashTablePtr createHashTable(int size) {

    hashTablePtr t;

    t = (hashTablePtr) malloc(sizeof(struct hashTable));
    t->size = size;
    t->data = (hashTableEntryPtr *) malloc(sizeof(hashTableEntryPtr)*size);
    memset(t->data, 0, sizeof(hashTableEntryPtr)*size);

    return t; 
}

int insertHashTableEntry(hashTablePtr table, 
                         int (*hash)(void *, int), 
                         int (*compare)(void *, void *),
                         void *key, 
                         void *value) {

    int h;
    hashTableEntryPtr entry;

    h = hash(key, table->size);

    if (table->data[h] == NULL) {
        table->data[h] = (hashTableEntryPtr) 
                                          malloc(sizeof(struct hashTableEntry));
        memset(table->data[h], 0, sizeof(struct hashTableEntry));
        table->data[h]->key = key;
        table->data[h]->value = value;
    } else {
        entry = table->data[h];
        do {
            if (compare(key, entry->key))
                return PDUPLICATED_KEY;
            if (entry->next == NULL) {
                entry->next = (hashTableEntryPtr) 
                                          malloc(sizeof(struct hashTableEntry));
                memset(entry->next, 0, sizeof(struct hashTableEntry));
                entry->next->key = key;
                entry->next->value = value;
                entry = NULL;
            } else {
                entry = entry->next;
            } 
        } while (entry != NULL);
        
    }

    return POK;
}

int removeHashTableEntry(hashTablePtr table, 
                         int (*hash)(void *, int),
                         int (*compare)(void *, void *),
                         void *key) {

    int h;
    hashTableEntryPtr entry, temp;

    h = hash(key, table->size);

    if (table->data[h] == NULL)
        return PNOTFOUND;

    entry = table->data[h];

    if (compare(table->data[h]->key, key)) {
        table->data[h] = table->data[h]->next;
        free(entry);
        return POK;
    } 

    while(entry->next != NULL) {
        if (compare(entry->next->key, key)) {
            temp = entry->next;
            entry->next = entry->next->next;
            free(temp); 
            return POK;
        }
        entry = entry->next;
    }

    return PNOTFOUND;
}

int lookupHashTableEntry(hashTablePtr table, 
                         int (*hash)(void *, int),
                         int (*compare)(void *, void *),
                         void *key, 
                         void **value) {

    int h;
    hashTableEntryPtr entry;

    h = hash(key, table->size);

    entry = table->data[h];
    while (entry != NULL) {
        if (compare(key, entry->key)) {
            *value = entry->value;
            return POK;
        }
        entry = entry->next;
    }

    *value = NULL;
    return PNOTFOUND;
}

hashTableEntryPtr getHashTableEntryArray(hashTablePtr table) {

    int i;
    hashTableEntryPtr result;
    hashTableEntryPtr temp;
    hashTableEntryPtr buff;

    if (table == NULL)
        return NULL;

    result = NULL;

    for (i=0; i<table->size; i++) {
        temp = table->data[i];
        while (temp != NULL) {
            buff = (hashTableEntryPtr) malloc(sizeof(struct hashTableEntry));
            memcpy(buff, temp, sizeof(struct hashTableEntry));
            buff->next = result;
            result = buff;

            temp = temp->next;
        }
    }

    return result;
}


/*
 * Hash function taken from Aho, Sethi, Ullman "Compilers --- principles,
 * techniques and tools", pp. 436. Credited to P.J. Weinberger
 */
int strHash(void *key, int tableSize)
{   
    char *p;
    unsigned h = 0, g;

    for(p=key; *p != '\0'; p++) {
        h = ((h << 4) + (*p));
        if ((g = h&0xf0000000)) {
            h = h ^ (g >> 24);
            h = h ^ g;
        }
    }

    return(h % tableSize);
}

int intHash(void *key, int tableSize) {

    return((int)key % tableSize);  
}

int strCompare(void *str1, void *str2) {

    if (!strcmp((char *) str1, (char *) str2))
        return 1;
    else 
        return 0;
} 

int intCompare(void *i1, void *i2) {

    if (i1 == i2)
        return 1;
    else 
        return 0;
}

